@@ -33,6 +33,7 @@ std::string s_numpy_load_error;
 bool haveNumPy() {
   return s_numpy_load_error.empty();
 }
+
 bool requireNumPy() {
   if (!haveNumPy())
     stop("Required version of NumPy not available: " + s_numpy_load_error);
@@ -89,15 +90,17 @@ PyObject* r_object_capsule(SEXP object) {
 
 
 // helper class for ensuring decref of PyObject in the current scope
-template<typename T>
+template <typename T>
 class PyPtr {
+  
 public:
+  
   // attach on creation, decref on destruction
   PyPtr() : object_(NULL) {}
   explicit PyPtr(T* object) : object_(object) {}
   virtual ~PyPtr() {
     if (object_ != NULL)
-      Py_DecRef((PyObject*)object_);
+      Py_DecRef((PyObject*) object_);
   }
 
   operator T*() const { return object_; }
@@ -573,15 +576,6 @@ int scalar_list_type(PyObject* x) {
   return scalarType;
 }
 
-// convert a tuple to a character vector
-CharacterVector py_tuple_to_character(PyObject* tuple) {
-  Py_ssize_t len = PyTuple_Size(tuple);
-  CharacterVector vec(len);
-  for (Py_ssize_t i = 0; i<len; i++)
-    vec[i] = as_utf8_r_string(PyTuple_GetItem(tuple, i));
-  return vec;
-}
-
 void set_string_element(SEXP rArray, int i, PyObject* pyStr) {
   std::string str = as_std_string(pyStr);
   cetype_t ce = PyUnicode_Check(pyStr) ? CE_UTF8 : CE_NATIVE;
@@ -942,9 +936,211 @@ SEXP py_to_r(PyObject* x, bool convert) {
   }
 }
 
+// [[Rcpp::export]]
+SEXP py_get_formals(PyObjectRef func) {
+  
+  PyObjectPtr inspect(py_import("inspect"));
+  if (inspect.is_null()) stop(py_fetch_error());
+  PyObjectPtr get_signature(PyObject_GetAttrString(inspect.get(), "signature"));
+  if (get_signature.is_null()) stop(py_fetch_error());
+  PyObjectPtr signature(PyObject_CallFunctionObjArgs(get_signature.get(), func.get(), NULL));
+  if (signature.is_null()) stop(py_fetch_error());
+  PyObjectPtr param_dict(PyObject_GetAttrString(signature.get(), "parameters"));
+  if (param_dict.is_null()) stop(py_fetch_error());
+
+  PyObjectPtr param_values(PyObject_GetAttrString(param_dict.get(), "values"));
+  if (param_values.is_null()) stop(py_fetch_error());
+  PyObjectPtr params(PyObject_CallFunctionObjArgs(param_values.get(), NULL, NULL));
+  if (params.is_null()) stop(py_fetch_error());
+  PyObjectPtr param_iter(PyObject_GetIter(params.get()));
+  if (param_iter.is_null()) stop(py_fetch_error());
+
+  // Static properties of the Parameter class
+  PyObjectPtr param_class(PyObject_GetAttrString(inspect.get(), "Parameter"));
+  if (param_class.is_null()) stop(py_fetch_error());
+  PyObjectPtr empty_param(PyObject_GetAttrString(param_class.get(), "empty"));
+  if (empty_param.is_null()) stop(py_fetch_error());
+  PyObjectPtr var_pos(PyObject_GetAttrString(param_class.get(), "VAR_POSITIONAL"));
+  if (var_pos.is_null()) stop(py_fetch_error());
+  PyObjectPtr var_kw(PyObject_GetAttrString(param_class.get(), "VAR_KEYWORD"));
+  if (var_kw.is_null()) stop(py_fetch_error());
+  PyObjectPtr kw_only(PyObject_GetAttrString(param_class.get(), "KEYWORD_ONLY"));
+  if (kw_only.is_null()) stop(py_fetch_error());
+
+  Rcpp::Pairlist formals;
+  bool var_encountered = false;
+  while (true) {
+    PyObjectPtr param(PyIter_Next(param_iter.get()));
+    if (param.is_null()) break;
+
+    PyObjectPtr param_name(PyObject_GetAttrString(param.get(), "name"));
+    if (param_name.is_null()) stop(py_fetch_error());
+    PyObjectPtr param_kind(PyObject_GetAttrString(param.get(), "kind"));
+    if (param_kind.is_null()) stop(py_fetch_error());
+    PyObjectPtr param_default(PyObject_GetAttrString(param.get(), "default"));
+    if (param_default.is_null()) stop(py_fetch_error());
+
+    // If we encounter our first kw_only param
+    // without having encountered `*args` or `**kw`,
+    // we insert `...` before the actual parameter.
+    if (param_kind == kw_only && !var_encountered) {
+      formals << Named("...", R_MissingArg);
+      var_encountered = true;
+    }
+
+    // If we encounter the first of `*args` or `**kw`,
+    // we insert `...` instead of a parameter.
+    // foo(*args, b=1, **kw) -> foo(..., b=1)
+    if (param_kind == var_pos || param_kind == var_kw) {
+      if (!var_encountered) {
+        formals << Named("...", R_MissingArg);
+        var_encountered = true;
+      }
+    // For a parameter w/o default value, we insert `R_MissingArg`.
+    // There is inspect.Parameter(..., *, default = Parameter.empty, ...)
+    // so we check if param_kind != kw_only for this corner case.
+    } else if (param_kind != kw_only && param_default == empty_param) {
+      formals << Named(as_utf8_r_string(param_name.get()), R_MissingArg);
+    // If we arrive here we have a parameter with default value.
+    } else {
+      // Here we could convert a subset of python objects to R defaults.
+      // Plain values (numeric, character, NULL, ...) are stored as is,
+      // variables, calls, ... are stored as `symbol` or `language`.
+      formals << Named(as_utf8_r_string(param_name.get()), R_NilValue);
+    }
+  }
+
+  return formals;
+  
+}
+
+bool is_convertible_to_numpy(RObject x) {
+  
+  if (!haveNumPy())
+    return false;
+  
+  int type = TYPEOF(x);
+  
+  return
+    type == INTSXP  ||
+    type == REALSXP ||
+    type == LGLSXP  ||
+    type == CPLXSXP ||
+    type == STRSXP;
+}
+
+PyObject* r_to_py_numpy(RObject x, bool convert) {
+  
+  int type = x.sexp_type();
+  SEXP sexp = x.get__();
+  
+  // figure out dimensions for resulting array
+  IntegerVector dimensions = x.hasAttribute("dim")
+    ? x.attr("dim")
+    : IntegerVector::create(Rf_xlength(x));
+  
+  int nd = dimensions.length();
+  std::vector<npy_intp> dims(nd);
+  for (int i = 0; i < nd; i++)
+    dims[i] = dimensions[i];
+  
+  // get pointer + type for underlying data
+  int typenum;
+  void* data;
+  if (type == INTSXP) {
+    if (sizeof(long) == 4)
+      typenum = NPY_LONG;
+    else
+      typenum = NPY_INT;
+    data = &(INTEGER(sexp)[0]);
+  } else if (type == REALSXP) {
+    typenum = NPY_DOUBLE;
+    data = &(REAL(sexp)[0]);
+  } else if (type == LGLSXP) {
+    typenum = NPY_BOOL;
+    data = &(LOGICAL(sexp)[0]);
+  } else if (type == CPLXSXP) {
+    typenum = NPY_CDOUBLE;
+    data = &(COMPLEX(sexp)[0]);
+  } else if (type == STRSXP) {
+    typenum = NPY_OBJECT;
+    data = NULL;
+  } else {
+    stop("Matrix type cannot be converted to python (only integer, "
+           "numeric, complex, logical, and character matrixes can be "
+           "converted");
+  }
+  
+  int flags = NPY_ARRAY_FARRAY_RO;
+  
+  // because R logical vectors are just ints under the
+  // hood, we need to explicitly construct a boolean
+  // vector for our Python array. note that the created
+  // array will own the data so we do not free it after
+  if (typenum == NPY_BOOL) {
+    R_xlen_t n = XLENGTH(sexp);
+    bool* converted = (bool*) PyArray_malloc(n * sizeof(bool));
+    for (R_xlen_t i = 0; i < n; i++)
+      converted[i] = LOGICAL(sexp)[i];
+    data = converted;
+    flags |= NPY_ARRAY_OWNDATA;
+  }
+  
+  // create the matrix
+  PyObject* array = PyArray_New(&PyArray_Type,
+                                nd,
+                                &(dims[0]),
+                                typenum,
+                                NULL,
+                                data,
+                                0,
+                                flags,
+                                NULL);
+  
+  // check for error
+  if (array == NULL)
+    stop(py_fetch_error());
+  
+  // if this is a character vector we need to convert and set the elements,
+  // otherwise the memory is shared with the underlying R vector
+  if (type == STRSXP) {
+    void** pData = (void**)PyArray_DATA((PyArrayObject*)array);
+    R_xlen_t len = Rf_xlength(x);
+    for (R_xlen_t i = 0; i<len; i++) {
+      PyObject* pyStr = as_python_str(STRING_ELT(x, i));
+      pData[i] = pyStr;
+    }
+    
+  } else {
+    // wrap the R object in a capsule that's tied to the lifetime of the matrix
+    // (so the R doesn't deallocate the memory while python is still pointing to it)
+    PyObjectPtr capsule(r_object_capsule(x));
+    
+    // set base object using correct version of the API (detach since this
+    // effectively steals a reference to the provided base object)
+    if (PyArray_GetNDArrayCFeatureVersion() >= NPY_1_7_API_VERSION) {
+      int res = PyArray_SetBaseObject((PyArrayObject *)array, capsule.detach());
+      if (res != 0)
+        stop(py_fetch_error());
+    } else {
+      PyArray_BASE(array) = capsule.detach();
+    }
+  }
+  
+  // return it
+  return array;
+  
+}
+
+PyObject* r_to_py_cpp(RObject x, bool convert);
+
 PyObject* r_to_py(RObject x, bool convert) {
 
-  // get a static reference to the R version of r_to_py
+  // if the object bit is not set, we can skip R dispatch
+  if (OBJECT(x) == 0)
+    return r_to_py_cpp(x, convert);
+  
+  // get a reference to the R version of r_to_py
   Rcpp::Environment pkgEnv = Rcpp::Environment::namespace_env("reticulate");
   Rcpp::Function r_to_py_fn = pkgEnv["r_to_py"];
 
@@ -953,7 +1149,7 @@ PyObject* r_to_py(RObject x, bool convert) {
   PyObjectRef ref(r_to_py_fn(x, convert));
 
   // get the underlying Python object and call Py_IncRef before returning it
-  // this allows this function to provide the same memory sematnics as the
+  // this allows this function to provide the same memory semantics as the
   // previous C++ version of r_to_py (which is now r_to_py_cpp), which always
   // calls Py_IncRef on Python objects before returning them
   PyObject* obj = ref.get();
@@ -963,6 +1159,21 @@ PyObject* r_to_py(RObject x, bool convert) {
   return obj;
 }
 
+// Python capsule wrapping an R's external pointer object
+static void free_r_extptr_capsule(PyObject* capsule) {
+  SEXP sexp = (SEXP)PyCapsule_GetContext(capsule);
+  ::R_ReleaseObject(sexp);
+}
+static PyObject* r_extptr_capsule(SEXP sexp) {
+  // underlying pointer
+  void* ptr = R_ExternalPtrAddr(sexp);
+  if (ptr == NULL) stop("Invalid pointer");
+  ::R_PreserveObject(sexp);
+  PyObject* capsule = PyCapsule_New(ptr, NULL, free_r_extptr_capsule);
+  PyCapsule_SetContext(capsule, (void*)sexp);
+  return capsule;
+}
+
 // convert an R object to a python object (the returned object
 // will have an active reference count on it)
 PyObject* r_to_py_cpp(RObject x, bool convert) {
@@ -992,97 +1203,7 @@ PyObject* r_to_py_cpp(RObject x, bool convert) {
 
   // convert arrays and matrixes to numpy (throw error if numpy not available)
   } else if (x.hasAttribute("dim") && requireNumPy()) {
-
-    IntegerVector dimAttrib = x.attr("dim");
-    int nd = dimAttrib.length();
-    std::vector<npy_intp> dims(nd);
-    for (int i = 0; i<nd; i++)
-      dims[i] = dimAttrib[i];
-    int typenum;
-    void* data;
-    if (type == INTSXP) {
-      if (sizeof(long) == 4)
-        typenum = NPY_LONG;
-      else
-        typenum = NPY_INT;
-      data = &(INTEGER(sexp)[0]);
-    } else if (type == REALSXP) {
-      typenum = NPY_DOUBLE;
-      data = &(REAL(sexp)[0]);
-    } else if (type == LGLSXP) {
-      typenum = NPY_BOOL;
-      data = &(LOGICAL(sexp)[0]);
-    } else if (type == CPLXSXP) {
-      typenum = NPY_CDOUBLE;
-      data = &(COMPLEX(sexp)[0]);
-    } else if (type == STRSXP) {
-      typenum = NPY_OBJECT;
-      data = NULL;
-    } else {
-      stop("Matrix type cannot be converted to python (only integer, "
-           "numeric, complex, logical, and character matrixes can be "
-           "converted");
-    }
-
-    int flags = NPY_ARRAY_FARRAY_RO;
-
-    // because R logical vectors are just ints under the
-    // hood, we need to explicitly construct a boolean
-    // vector for our Python array. note that the created
-    // array will own the data so we do not free it after
-    if (typenum == NPY_BOOL) {
-      R_xlen_t n = XLENGTH(sexp);
-      bool* converted = (bool*) PyArray_malloc(n * sizeof(bool));
-      for (R_xlen_t i = 0; i < n; i++)
-        converted[i] = LOGICAL(sexp)[i];
-      data = converted;
-      flags |= NPY_ARRAY_OWNDATA;
-    }
-
-    // create the matrix
-    PyObject* array = PyArray_New(&PyArray_Type,
-                                   nd,
-                                   &(dims[0]),
-                                   typenum,
-                                   NULL,
-                                   data,
-                                   0,
-                                   flags,
-                                   NULL);
-
-    // check for error
-    if (array == NULL)
-      stop(py_fetch_error());
-
-    // if this is a character vector we need to convert and set the elements,
-    // otherwise the memory is shared with the underlying R vector
-    if (type == STRSXP) {
-      void** pData = (void**)PyArray_DATA((PyArrayObject*)array);
-      R_xlen_t len = Rf_xlength(x);
-      for (R_xlen_t i = 0; i<len; i++) {
-        PyObject* pyStr = as_python_str(STRING_ELT(x, i));
-        pData[i] = pyStr;
-      }
-
-    } else {
-      // wrap the R object in a capsule that's tied to the lifetime of the matrix
-      // (so the R doesn't deallocate the memory while python is still pointing to it)
-      PyObjectPtr capsule(r_object_capsule(x));
-
-      // set base object using correct version of the API (detach since this
-      // effectively steals a reference to the provided base object)
-      if (PyArray_GetNDArrayCFeatureVersion() >= NPY_1_7_API_VERSION) {
-        int res = PyArray_SetBaseObject((PyArrayObject *)array, capsule.detach());
-        if (res != 0)
-          stop(py_fetch_error());
-      } else {
-        PyArray_BASE(array) = capsule.detach();
-      }
-    }
-
-    // return it
-    return array;
-
+    return r_to_py_numpy(x, convert);
   // integer (pass length 1 vectors as scalars, otherwise pass list)
   } else if (type == INTSXP) {
     if (LENGTH(sexp) == 1) {
@@ -1230,6 +1351,9 @@ PyObject* r_to_py_cpp(RObject x, bool convert) {
     // return the wrapper
     return wrapper.detach();
 
+  // externalptr
+  } else if (type == EXTPTRSXP) {
+    return r_extptr_capsule(sexp);
   } else {
     Rcpp::print(sexp);
     stop("Unable to convert R object to Python type");
@@ -1250,6 +1374,7 @@ extern "C" PyObject* call_r_function(PyObject *self, PyObject* args, PyObject* k
   // the first argument is always the capsule containing the R function to call
   PyObject* capsule = PyTuple_GetItem(args, 0);
   RObject rFunction = r_object_from_capsule(capsule);
+
   bool convert = (bool)PyCapsule_GetContext(capsule);
 
   // convert remainder of positional arguments to R list
@@ -1268,15 +1393,18 @@ extern "C" PyObject* call_r_function(PyObject *self, PyObject* args, PyObject* k
 
   // get keyword arguments
   List rKeywords;
-  if (convert) {
-    rKeywords = ::py_to_r(keywords, convert);
-  } else {
-    PyObject *key, *value;
-    Py_ssize_t pos = 0;
-    while (PyDict_Next(keywords, &pos, &key, &value)) {
-      PyObjectPtr str(PyObject_Str(key));
-      Py_IncRef(value);
-      rKeywords[as_std_string(str)] = py_ref(value, convert);
+
+  if (keywords != NULL) {
+    if (convert) {
+      rKeywords = ::py_to_r(keywords, convert);
+    } else {
+      PyObject *key, *value;
+      Py_ssize_t pos = 0;
+      while (PyDict_Next(keywords, &pos, &key, &value)) {
+        PyObjectPtr str(PyObject_Str(key));
+        Py_IncRef(value);
+        rKeywords[as_std_string(str)] = py_ref(value, convert);
+      }
     }
   }
 
@@ -1587,6 +1715,7 @@ void py_initialize(const std::string& python,
 
   // poll for events while executing python code
   event_loop::initialize();
+
 }
 
 // [[Rcpp::export]]
@@ -1720,36 +1849,71 @@ bool py_has_attr_impl(PyObjectRef x, const std::string& name) {
     return PyObject_HasAttrString(x, name.c_str());
 }
 
-// [[Rcpp::export]]
-PyObjectRef py_get_attr_impl(PyObjectRef x, const std::string& name, bool silent = false) {
+namespace {
 
-  PyObject* attr = PyObject_GetAttrString(x, name.c_str());
-
-  if (attr == NULL) {
-
-    std::string err = py_fetch_error();
-
-    if (!silent) {
-      // error if we aren't silent
-      stop(err);
-    } else {
-      // otherwise set it to PyNone
-      attr = Py_None;
-      Py_IncRef(attr);
-    }
+PyObjectRef py_get_common(PyObject* object,
+                          bool convert,
+                          bool silent)
+{
+  // if we have an object, return it
+  if (object != NULL)
+    return py_ref(object, convert);
+  
+  // if we're silent, return Py_None
+  if (silent) {
+    Py_IncRef(Py_None);
+    return py_ref(Py_None, convert);
   }
+  
+  // otherwise, throw an R error
+  stop(py_fetch_error());
+  
+}
+  
+} // end anonymous namespace
 
-  return py_ref(attr, x.convert());
+// [[Rcpp::export]]
+PyObjectRef py_get_attr_impl(PyObjectRef x,
+                             const std::string& key,
+                             bool silent = false)
+{
+  PyObject* attr = PyObject_GetAttrString(x, key.c_str());
+  return py_get_common(attr, x.convert(), silent);
 }
 
 // [[Rcpp::export]]
-void py_set_attr_impl(PyObjectRef x, const std::string& name, RObject value) {
+PyObjectRef py_get_item_impl(PyObjectRef x,
+                             RObject key,
+                             bool silent = false)
+{
+  PyObjectPtr py_key(r_to_py(key, x.convert()));
+  PyObject* item = PyObject_GetItem(x, py_key);
+  return py_get_common(item, x.convert(), silent);
+}
+
+// [[Rcpp::export]]
+void py_set_attr_impl(PyObjectRef x,
+                      const std::string& name,
+                      RObject value)
+{
   PyObjectPtr converted(r_to_py(value, x.convert()));
   int res = PyObject_SetAttrString(x, name.c_str(), converted);
   if (res != 0)
     stop(py_fetch_error());
 }
 
+// [[Rcpp::export]]
+void py_set_item_impl(PyObjectRef x,
+                      RObject key,
+                      RObject val)
+{
+  PyObjectPtr py_key(r_to_py(key, x.convert()));
+  PyObjectPtr py_val(r_to_py(val, x.convert()));
+  
+  int res = PyObject_SetItem(x, py_key, py_val);
+  if (res != 0)
+    stop(py_fetch_error());
+}
 
 
 // [[Rcpp::export]]
@@ -1849,22 +2013,31 @@ SEXP py_call_impl(PyObjectRef x, List args = R_NilValue, List keywords = R_NilVa
   return py_ref(res.detach(), x.convert());
 }
 
-
 // [[Rcpp::export]]
 PyObjectRef py_dict_impl(const List& keys, const List& items, bool convert) {
+  
   PyObject* dict = PyDict_New();
-  for (R_xlen_t i = 0; i<keys.length(); i++) {
+  
+  for (R_xlen_t i = 0; i < keys.length(); i++) {
     PyObjectPtr key(r_to_py(keys.at(i), convert));
-    PyObjectPtr item(r_to_py(items.at(i), convert));
-    PyDict_SetItem(dict, key, item);
+    PyObjectPtr val(r_to_py(items.at(i), convert));
+    PyDict_SetItem(dict, key, val);
   }
+  
   return py_ref(dict, convert);
+  
 }
 
 
 // [[Rcpp::export]]
 SEXP py_dict_get_item(PyObjectRef dict, RObject key) {
+  
+  if (!PyDict_Check(dict))
+    return py_get_item_impl(dict, key, false);
+  
   PyObjectPtr pyKey(r_to_py(key, dict.convert()));
+  
+  // NOTE: returns borrowed reference
   PyObject* item = PyDict_GetItem(dict, pyKey);
   if (item != NULL) {
     Py_IncRef(item);
@@ -1873,69 +2046,124 @@ SEXP py_dict_get_item(PyObjectRef dict, RObject key) {
     Py_IncRef(Py_None);
     return py_ref(Py_None, false);
   }
+  
 }
 
 // [[Rcpp::export]]
-void py_dict_set_item(PyObjectRef dict, RObject item, RObject value) {
-  PyObjectPtr pyItem(r_to_py(item, dict.convert()));
-  PyObjectPtr pyValue(r_to_py(value, dict.convert()));
-  PyDict_SetItem(dict, pyItem, pyValue);
+void py_dict_set_item(PyObjectRef dict, RObject key, RObject val) {
+  
+  if (!PyDict_Check(dict))
+    return py_set_item_impl(dict, key, val);
+  
+  PyObjectPtr py_key(r_to_py(key, dict.convert()));
+  PyObjectPtr py_val(r_to_py(val, dict.convert()));
+  PyDict_SetItem(dict, py_key, py_val);
+  
 }
 
 // [[Rcpp::export]]
 int py_dict_length(PyObjectRef dict) {
+  
+  if (!PyDict_Check(dict))
+    return PyObject_Size(dict);
+  
   return PyDict_Size(dict);
+  
 }
 
+namespace {
+
+PyObject* py_dict_get_keys_impl(PyObject* dict) {
+  
+  PyObject* keys = PyDict_Keys(dict);
+  
+  if (keys == NULL) {
+    PyErr_Clear();
+    keys = PyObject_CallMethod(dict, "keys", NULL);
+    if (keys == NULL)
+      stop(py_fetch_error());
+  }
+  
+  return keys;
+  
+}
+
+} // end anonymous namespace
+
 // [[Rcpp::export]]
 PyObjectRef py_dict_get_keys(PyObjectRef dict) {
-  return py_ref(PyDict_Keys(dict), dict.convert());
+  PyObject* keys = py_dict_get_keys_impl(dict);
+  return py_ref(keys, dict.convert());
 }
 
 // [[Rcpp::export]]
 CharacterVector py_dict_get_keys_as_str(PyObjectRef dict) {
 
-  // get the keys and check their length
-  PyObjectPtr pyKeys(PyDict_Keys(dict));
-  Py_ssize_t len = PyList_Size(pyKeys);
+  // get the dictionary keys
+  PyObjectPtr py_keys(py_dict_get_keys_impl(dict));
+  
+  // iterate over keys and convert to string
+  std::vector<std::string> keys;
 
-  // allocate keys to return
-  CharacterVector keys(len);
-
-  // get the keys as strings
-  for (Py_ssize_t i = 0; i<len; i++) {
-    PyObject* item = PyList_GetItem(pyKeys, i);
+  PyObjectPtr it(PyObject_GetIter(py_keys));
+  if (it == NULL)
+    stop(py_fetch_error());
+  
+  for (PyObject* item = PyIter_Next(it);
+       item != NULL;
+       item = PyIter_Next(it))
+  {
+    // decref on scope exit
+    PyObjectPtr scope(item);
+    
+    // check for python string and use directly
     if (is_python_str(item)) {
-      keys[i] = as_utf8_r_string(item);
-    } else {
-      PyObjectPtr str(PyObject_Str(item));
-      if (str.is_null())
-        stop(py_fetch_error());
-      keys[i] = as_utf8_r_string(str);
+      keys.push_back(as_utf8_r_string(item));
+      continue;
     }
+    
+    // if we don't have a python string, try to create one
+    PyObjectPtr str(PyObject_Str(item));
+    if (str.is_null())
+      stop(py_fetch_error());
+    
+    keys.push_back(as_utf8_r_string(str));
+    
   }
-
-  // return
-  return keys;
+  
+  if (PyErr_Occurred())
+    stop(py_fetch_error());
+  
+  return CharacterVector(keys.begin(), keys.end());
+  
 }
 
 
 // [[Rcpp::export]]
 PyObjectRef py_tuple(const List& items, bool convert) {
-  PyObject* tuple = PyTuple_New(items.length());
-  for (R_xlen_t i = 0; i<items.length(); i++) {
+  
+  R_xlen_t n = items.length();
+  PyObject* tuple = PyTuple_New(n);
+  for (R_xlen_t i = 0; i < n; i++) {
     PyObject* item = r_to_py(items.at(i), convert);
     // NOTE: reference to arg is "stolen" by the tuple
     int res = PyTuple_SetItem(tuple, i, item);
     if (res != 0)
       stop(py_fetch_error());
   }
+  
   return py_ref(tuple, convert);
+  
 }
 
 // [[Rcpp::export]]
 int py_tuple_length(PyObjectRef tuple) {
+  
+  if (!PyTuple_Check(tuple))
+    return PyObject_Size(tuple);
+  
   return PyTuple_Size(tuple);
+  
 }
 
 
@@ -2107,10 +2335,16 @@ SEXP py_run_file_impl(const std::string& file,
 SEXP py_eval_impl(const std::string& code, bool convert = true) {
 
   // compile the code
-  PyObjectPtr compiledCode(Py_CompileString(code.c_str(), "reticulate_eval", Py_eval_input));
+  PyObjectPtr compiledCode;
+  if (Py_CompileStringExFlags != NULL) 
+    compiledCode.assign(Py_CompileStringExFlags(code.c_str(), "reticulate_eval", Py_eval_input, NULL, 0));
+  else 
+    compiledCode.assign(Py_CompileString(code.c_str(), "reticulate_eval", Py_eval_input));
+  
+  
   if (compiledCode.is_null())
     stop(py_fetch_error());
-
+  
   // execute the code
   PyObject* main = PyImport_AddModule("__main__");
   PyObject* dict = PyModule_GetDict(main);
@@ -2127,3 +2361,222 @@ SEXP py_eval_impl(const std::string& code, bool convert = true) {
   return result;
 
 }
+
+// [[Rcpp::export]]
+SEXP py_convert_pandas_series(PyObjectRef series) {
+  
+  // extract dtype
+  PyObjectPtr dtype(PyObject_GetAttrString(series, "dtype"));
+  PyObjectPtr name(PyObject_GetAttrString(dtype, "name"));
+  
+  RObject R_obj;
+  
+  // special treatment for pd.Categorical
+  if (as_std_string(name) == "category") {
+    
+    // get actual values and convert to R
+    PyObjectPtr cat(PyObject_GetAttrString(series, "cat"));
+    PyObjectPtr codes(PyObject_GetAttrString(cat, "codes"));
+    PyObjectPtr code_values(PyObject_GetAttrString(codes, "values"));
+    RObject R_values = py_to_r(code_values, true);
+    
+    // get levels and convert to R
+    PyObjectPtr categories(PyObject_GetAttrString(dtype, "categories"));
+    PyObjectPtr category_values(PyObject_GetAttrString(categories, "values"));
+    RObject R_levels = py_to_r(category_values, true);
+    
+    // get "ordered" attribute
+    PyObjectPtr ordered(PyObject_GetAttrString(dtype, "ordered"));
+    //RObject ordered = py_to_r(ordered_, true);
+    
+    // populate integer vector to hold factor values
+    int* codes_int = INTEGER(R_values);
+    int n = Rf_xlength(R_values);
+    IntegerVector factor(n);
+    // values need to start at 1
+    for (int i = 0; i < n; ++i) {
+      factor[i] = codes_int[i] + 1;
+    }
+    
+    // populate character vector to hold levels
+    CharacterVector factor_levels(R_levels);
+    factor_levels.attr("dim") = R_NilValue;
+    
+    factor.attr("class") = "factor";
+    factor.attr("levels") = factor_levels;
+    if (PyObject_IsTrue(ordered)) factor.attr("ordered") = true;
+    
+    R_obj = factor;
+    
+  // special treatment for pd.TimeStamp
+  // if available, time zone information will be respected,
+  // but values returned to R will be in UTC
+  } else if (as_std_string(name) == "datetime64[ns]" ||
+    // if a time zone is present, dtype is "object"
+    PyObject_HasAttrString(series, "dt")) {
+
+    // pd.Series.items() returns an iterator over (index, value) pairs
+    PyObjectPtr items(PyObject_CallMethod(series, "items", NULL));
+    
+    std::vector<double> posixct;
+    
+    while (true) {
+      
+      // get next tuple
+      PyObjectPtr tuple(PyIter_Next(items));
+      if (tuple.is_null()) {
+        if (PyErr_Occurred())
+          stop(py_fetch_error());
+        else
+          break;
+      }
+    
+     // access value in slot 1
+     PyObjectPtr values(PySequence_GetItem(tuple, 1));
+     // convert to POSIX timestamp, taking into account time zone (if set)
+     PyObjectPtr timestamp(PyObject_CallMethod(values, "timestamp", NULL));
+     
+     Datetime R_timestamp;
+     
+     // NaT will have thrown "NaTType does not support timestamp"
+     if (PyErr_Occurred()) {
+       R_timestamp = R_NaN;
+       PyErr_Clear();
+     } else {
+       R_timestamp = py_to_r(timestamp, true);
+     }
+     
+     posixct.push_back(R_timestamp);
+    }
+    
+    DatetimeVector R_posixct(posixct.size());
+    for (std::size_t i = 0; i < posixct.size(); ++i) {
+      R_posixct[i] = posixct[i];
+    }
+    
+    return R_posixct;
+    
+  // default case
+  } else {
+    
+    PyObjectPtr values(PyObject_GetAttrString(series, "values"));
+    R_obj = py_to_r(values, series.convert());
+    
+  }
+  
+  return R_obj;
+  
+}
+
+// [[Rcpp::export]]
+SEXP py_convert_pandas_df(PyObjectRef df) {
+  
+  // pd.DataFrame.items() returns an iterator over (column name, Series) pairs
+  PyObjectPtr items(PyObject_CallMethod(df, "items", NULL));
+  if (! (PyObject_HasAttrString(items, "__next__") || PyObject_HasAttrString(items, "next")))
+    stop("Cannot iterate over object");
+
+  std::vector<RObject> list;
+  
+  while (true) {
+
+    // get next tuple
+    PyObjectPtr tuple(PyIter_Next(items));
+    if (tuple.is_null()) {
+      if (PyErr_Occurred())
+        stop(py_fetch_error());
+      else
+        break;
+    }
+    
+    // access Series in slot 1
+    PyObjectPtr series(PySequence_GetItem(tuple, 1));
+    // delegate to py_convert_pandas_series
+    PyObjectRef series_ref(series.detach(), df.convert());
+    RObject R_obj = py_convert_pandas_series(series_ref);
+
+    list.push_back(R_obj);
+  }
+
+  List rList(list.begin(), list.end());
+  return rList;
+
+}
+
+// [[Rcpp::export]]
+PyObjectRef r_convert_dataframe(RObject dataframe, bool convert) {
+  
+  Function r_convert_dataframe_column =
+    Environment::namespace_env("reticulate")["r_convert_dataframe_column"];
+  
+  PyObjectPtr dict(PyDict_New());
+  
+  CharacterVector names = dataframe.attr("names");
+  for (R_xlen_t i = 0, n = Rf_xlength(dataframe); i < n; i++)
+  {
+    RObject column = VECTOR_ELT(dataframe, i);
+    
+    int status = 0;
+    if (OBJECT(column) == 0) {
+      if (is_convertible_to_numpy(column)) {
+        PyObjectPtr value(r_to_py_numpy(column, convert));
+        status = PyDict_SetItemString(dict, names[i], value);
+      } else {
+        PyObjectPtr value(r_to_py_cpp(column, convert));
+        status = PyDict_SetItemString(dict, names[i], value);
+      }
+    } else {
+      PyObjectRef ref(r_convert_dataframe_column(column, convert));
+      status = PyDict_SetItemString(dict, names[i], ref.get());
+    }
+    
+    if (status != 0)
+      stop(py_fetch_error());
+  }
+  
+  return py_ref(dict.detach(), convert);
+
+}
+
+namespace {
+
+PyObject* r_convert_date_impl(PyObject* datetime,
+                              Date date)
+{
+  PyObjectPtr py_date(PyObject_CallMethod(
+      datetime, "date", "iii",
+      static_cast<int>(date.getYear()),
+      static_cast<int>(date.getMonth()),
+      static_cast<int>(date.getDay())));
+  
+  if (py_date == NULL)
+    stop(py_fetch_error());
+  
+  return py_date.detach();
+}
+
+} // end anonymous namespace
+
+// [[Rcpp::export]]
+PyObjectRef r_convert_date(DateVector dates, bool convert) {
+  
+  PyObjectPtr datetime(PyImport_ImportModule("datetime"));
+  
+  // short path for n == 1
+  R_xlen_t n = dates.size();
+  if (n == 1) {
+    Date date = dates[0];
+    return py_ref(r_convert_date_impl(datetime, date), convert);
+  }
+  
+  // regular path for n > 1
+  PyObjectPtr list(PyList_New(n));
+  
+  for (R_xlen_t i = 0; i < n; ++i) {
+    Date date = dates[i];
+    PyList_SetItem(list, i, r_convert_date_impl(datetime, date));
+  }
+  
+  return py_ref(list.detach(), convert);
+  
+}
